/* Copyright (C) 1999 Beau Kuiper

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include "ftpd.h"

void file_chuser(uid_t uid, uid_t gid)
{
	if (config->rootmode)
	{
#if DEBUG
		printf("[moving to uid: %d gid: %d]\n", uid, gid);
#endif

#if (HAVE_SETFSGID == 1)
		setfsgid(gid);
#else
		setegid(gid);
#endif
#if (HAVE_SETFSUID == 1)
		setfsuid(uid);
#else
		seteuid(uid);
#endif
	}
}

int checkdir(char *dir)
{
	struct stat check;
	int result = stat(dir, &check);

	if (result == 0)
		if (S_ISDIR(check.st_mode))
			return(TRUE);

	return(FALSE);
}

int checkfile(char *file)
{
	struct stat check;
	int result = stat(file, &check);

	if (result == 0)
		if (!S_ISDIR(check.st_mode))
			return(TRUE);
	return(FALSE);
}

int checkchdir(FTPSTATE *peer, char *dir)
{
	int result;
	file_chuser(peer->asuid, peer->asgid);
	result = (chdir(dir) == 0);
	file_chuser(0,0);
	return(result);
}

/* this needs explaining, send this function a pointer to nine '-' in a string ie "---------"
   and the file mode of the file, and it will send back the permissions for that file in that
   string, eg "-rw-r--r--" */

/* this version was inspired by how GNU fileutils does it */
   
void convertperms(char *permstr, int mode)
{
	int mcpy = mode;
	int cnt;
	
	/* work out the 3 sets of rwx's */
	
	for(cnt = 6; cnt > -1; cnt -= 3)
	{
		permstr[cnt+1] = (mcpy & S_IROTH) ? 'r' : '-';
		permstr[cnt+2] = (mcpy & S_IWOTH) ? 'w' : '-';
		permstr[cnt+3] = (mcpy & S_IXOTH) ? 'x' : '-';
		mcpy = (mcpy >> 3);
	}

	/* work out special bits. Setuid the setgid the sticky bit */
		
	if (mode & S_ISUID) permstr[3] = (mode & S_IXUSR) ? 's' : 'S';
	if (mode & S_ISGID) permstr[6] = (mode & S_IXGRP) ? 's' : 'S';
	if (mode & S_ISVTX) permstr[9] = (mode & S_IXOTH) ? 't' : 'T';
	
	/* check for regular files first, better performace! */
	if (S_ISREG(mode))
		permstr[0] = '-'; 
	else if (S_ISLNK(mode))
		permstr[0] = 'l';
	else if (S_ISDIR(mode))
		permstr[0] = 'd';
	else if (S_ISBLK(mode))
		permstr[0] = 'b';
	else if (S_ISCHR(mode))
		permstr[0] = 'c';
	else if (S_ISFIFO(mode))
		permstr[0] = 'p';
	else if (S_ISSOCK(mode))
		permstr[0] = 's';
	else
		permstr[0] = '?';
}

char *file_expand(FTPSTATE *peer, char *filename)
{
	char *newfile = strdupwrapper(peer->pwd);
	dir_combine(peer, &newfile, filename);
	return(newfile);
}

/* converts erron to MYERR constants */

int file_creat(FTPSTATE *peer, char *infile)
{
	char *filename = file_expand(peer, infile);
	int newfd = -1;
	
	if (checkfile(filename))
		check_acl(peer, filename, 'W');
	else 
		check_acl(peer, filename, 'A');
	
	if (errno == 0)
	{
		file_chuser(peer->asuid, peer->asgid);
		newfd = creat(filename, 0666 & ~peer->umask);
		file_chuser(0, 0);
	}
	
	freewrapper(filename);
	return(newfd);
}

int file_open(FTPSTATE *peer, char *infile, int flag)
{
	int filefd = -1;
	char *filename = file_expand(peer, infile);
	
	flag = (flag & (O_WRONLY | O_RDONLY));

	if (flag & O_WRONLY)
		check_acl(peer, filename, 'W');
	else
		check_acl(peer, filename, 'R');
	
	if (errno == 0)
		if (checkdir(filename))
			errno = EISDIR;
	
	if (errno == 0)
	{
		file_chuser(peer->asuid, peer->asgid);
		filefd = open(filename, flag);
		file_chuser(0, 0);
	}
	
	freewrapper(filename);
	return(filefd);
}

NEWFILE *file_nfopen(FTPSTATE *peer, char *filename)
{
	int fd = file_open(peer, filename, O_RDONLY);
	NEWFILE *result = NULL;
	
	if (fd > 0)
		result = nfdopen(fd);
	
	return(result);
} 

int file_glob(FTPSTATE *peer, char *infile, glob_t *gdata, int allfiles)
{
	int result = -1;
	char *dir, *file;
	
	/* escape all meta characters in directory name */
	
	dir = strdupwrapper(infile);
	*strrchr(dir, '/') = 0;
	file = safe_snprintf(".%s", strrchr(infile, '/'));

	check_acl(peer, infile, 'L');
	
	if (errno == 0)
	{
		file_chuser(peer->asuid, peer->asgid);
		result = chdir(dir);
		if (errno == 0)
		{
#ifdef GLOB_PERIOD
			if (allfiles)
				result = glob(file, GLOB_PERIOD, NULL, gdata);
			else
#endif
				result = glob(file, 0, NULL, gdata);
		}
		file_chuser(0, 0);
	}
	
	freewrapper(file);
	freewrapper(dir);
	return(result);
}

int file_unlink(FTPSTATE *peer, char *infile)
{
	int result = -1;
	char *filename = file_expand(peer, infile);

	check_acl(peer, filename, 'D');
	
	if (errno == 0)
	{
		file_chuser(peer->asuid, peer->asgid); 
		result = unlink(filename);
		file_chuser(0, 0);
	}
	
	freewrapper(filename);
	return(result);
}

char *file_rename(FTPSTATE *peer, char *infile, char *infile2)
{
	char *filename = file_expand(peer, infile);
	char *filename2 = file_expand(peer, infile2);
	char *badfile = NULL;
	
	if (checkfile(filename2))
		if (check_acl(peer, filename2, 'D') == -1)
			badfile = infile2;
 	if (errno == 0)
		if (check_acl(peer, filename2, 'A') == -1)
			badfile = infile2; 
	if (errno == 0)
		if (check_acl(peer, filename2, 'D') == -1)
			badfile = infile2;
	if (errno == 0)
		if (check_acl(peer, filename, 'R') == -1)
			badfile = infile;
		
	if (errno == 0)
	{
		file_chuser(peer->asuid, peer->asgid);
		if (rename(filename, filename2) == -1)
			badfile = infile;
		file_chuser(0, 0);
	}
	
	freewrapper(filename);
	freewrapper(filename2);
	return(badfile);
}

int file_stat(FTPSTATE *peer, char *infile, struct stat *output)
{
	int result = -1;
	char *filename = file_expand(peer, infile);
	
	check_acl(peer, filename, 'R');
	
	if (errno == 0)
	{
		file_chuser(peer->asuid, peer->asgid);
		result = stat(filename, output);
		file_chuser(0, 0);
	}
	
	freewrapper(filename);
	return(result);
}

int file_mkdir(FTPSTATE *peer, char *infile)
{
	int result = -1;
	char *filename = file_expand(peer, infile);

	check_acl(peer, filename, 'A');
	
	if (errno == 0)
	{
		file_chuser(peer->asuid, peer->asgid);
		result = mkdir(filename, 0777 & ~peer->umask);
		file_chuser(0, 0);
	}
	
	freewrapper(filename);
	return(result);
}

int file_rmdir(FTPSTATE *peer, char *infile)
{
	int result = -1;
	char *filename = file_expand(peer, infile);
	
	check_acl(peer, filename, 'D');
	
	if (errno == 0)
	{
		file_chuser(peer->asuid, peer->asgid);
		result = rmdir(filename);
		file_chuser(0, 0);
	}
	
	freewrapper(filename);
	return(result);
}

int file_chmod(FTPSTATE *peer, char *infile, int mode)
{
	int result = -1;
	char *filename = file_expand(peer, infile);
	
	check_acl(peer, filename, 'W');
	
	if (errno == 0)
	{
		file_chuser(peer->asuid, peer->asgid);
		
		/* never allow sticky bit or setgid or setuid to be set */
		
		result = chmod(filename, mode & 0777);
		file_chuser(0, 0);
	}
	
	freewrapper(filename);
	return(result);
}
