/* Copyright (C) 1999 Beau Kuiper

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

/* Includes */

#ifndef DEBUG		/* If we are not debuging, disable assertions! */
#define NDEBUG
#endif

#include "config.h"	/* Insert the config header */

#include <assert.h> 	/* enable assertions */

#if (HAVE_GLOB == 1)
#include <glob.h>	/* use the systems glob, faster and more accurate */
#else
#include "util/sglob.h"		/* use our simple glob */
#endif

#include "util/strerror.h"
#include <sys/utsname.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <pwd.h>
#include <grp.h>
#include <stdarg.h>
#include <sys/wait.h>
#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <string.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <time.h>
#include <signal.h>
#include <ctype.h>
#include <dirent.h>
#include <malloc.h>

#ifdef HAVE_CRYPT_H
#include <crypt.h>
#endif

#ifdef HAVE_GETSPNAM
#include <shadow.h>
#endif

#include "pnums.h"

/* Inportant Constants */

#define		FALSE		0
#define		TRUE		!FALSE

/* #define		LOGLEVELHIGH	LOG_CRIT */

#define		MAXPENDCONN	10	/* Maximum connections for listen */
#define		BUFFERSIZE	4096	/* Buffer size. Increase to improve
					   throughput performace */
#define		LOWATERMARK	4096	/* wait for this much data before
					   sending/receiving */
#define		BUFFSMALL	256	/* Filename and username buffer */
#define		STRCACHESIZE	20	/* Elements in a string cache */
#define		MAXLISTSIZE	262144  /* Maximum size for a list */
#define		MAXMSGLEN	16384	/* maximum size for a string */ 

#define		MYLOG_DACCESS	1
#define		MYLOG_FTRANS	2
#define		MYLOG_COMMAND	4
#define		MYLOG_RESPONSE	8
#define		MYLOG_INFO	16
#define 	MYLOG_LOGIN	32

/* DO NOT CHANGE THESE DEFINES. THEY ARE REQUIRED CONSTANTS */

#define CR			13
#define LF			10

#define PROGNAME		"muddleftpd"
#define VERSTR			"1.1.1a"
 
/* Macros */

#define MAXIMUM(x, y)		(x) > (y) ? (x) : (y)
#define MINIMUM(x, y)		(x) > (y) ? (y) : (x)

/* this specifies that the operation cannot be stopped by a signal */

#define NOSIGNALINTR(x)		while(((x) == -1) && (errno == EINTR))

/* Typedefs */

typedef struct {
	char *params[128];
	char *(* gettoken)(char, void *);
} TOKENSET;

typedef struct
{
	int fd;
	char *buffer;
	int bufferlen;
	int eof;
} NEWFILE;

typedef struct outbuff
{
	int fd;
	int size;
	char *outblock;
} OUTBUFF;

typedef struct strcache
{
	int size;
	struct {
		int num;
		char *str;
	       } data[STRCACHESIZE];
} STRCACHE;

typedef struct aclcacheent
{
	int pathlen;
	char *path;
	char *perms;
} ACLCACHEENT;

struct selector;

struct selectorobj
{
	int (* readsockopt)(struct selector *, void *);
	int (* writesockopt)(struct selector *, void *);
	void *readdata;
	void *writedata;
	int next;		/* linked list of file descriptors */
	int last;
};

typedef struct selector
{
	fd_set readset;
	fd_set writeset;
	struct selectorobj **fdtable;
	int firstfd;
	int maxfds;
	int smax;
} SELECTER;

typedef struct dataport
{
	int *fdout;		/* these get linked to either filefd or remotefd */ 
	int *fdin;
	int filefd;
	int remotefd;
	int pos;
	int startpos;
	int buffersize;
	int (* filter)(char *, int *);
	int (* firstop)(SELECTER *, void *);
	char *buffer;
	char *bufferpos;
	int die;
	int passive;
	int tsize;
	int graceful;
} DATAPORT;
	
typedef struct ftpstate
{
	int remotefd;		/* fd of the port to the remote system */
	unsigned int remoteip;		/* the systems IP number */
	int connport;		/* port connected to */
	char *username;		/* the username of the ftp user */
	char *pwd;		/* the current dir the ftp user is in */
	char *homedir;		/* the home directory of the ftp user */
	char *basedir;		/* the / directory for that user */
	int passiveport;	/* the fd of the listening passive port */
	char *renameoldname;	/* the source file to rename */
	int loggedin;		/* true if the user is logged in */
	int restartpos;		/* the position to start a download or upload */
	int remoteport;		/* the port number to connect to for data conns */
	int binary;		/* true if binary mode is set */
	char *logindump;	/* file to print after login */
	int threadnum;		/* the number of the thread in ftpd terms (not pid) */
	char *inbuffer;		/* input buffer */
	int inbuffersize;	/* input buffer size */
	char *cwddump;		/* dump this file when you move to this dir */
	char *busydump;		/* dump this file if too many users are logged in */
	char *quitdump;		/* dump this file when the user quits */
	ACLCACHEENT *acldata;	/* acl cache */
	int umask;
	gid_t asgid;		/* all file work is done as this gid (if possible) */
	uid_t asuid;		/* all file work is done as this uid (if possible) */
	int maxusers;		/* maximum instances of this user allowed to be logged in */
	int maxtimeout;		/* maximum time out for this user */
	int timeout;		/* timeout for this user */
	int chmodable;		/* can do chmod and change umask */
	SELECTER *sel;		/* select data stored here */
	DATAPORT *dport;	/* Data port information */
	int loginsleft;		/* number of login attempts left */
	int jailenabled;	/* is this user a jail */
	
	/* Counters for stats */
	int downloadedfiles;
	int uploadedfiles;
	int downloadedfilebytes;
	int uploadedfilebytes;
	int listdownloadedbytes;
	int listconns;
} FTPSTATE;

typedef struct ftpcmd
{
	char *command;
	int (* ftpfunc)(FTPSTATE *peer, char *params);
	int paramnum;
	int needslogin;	
	int dataportok;
	char *helpdesc;
} FTPCMD;

typedef struct inputline
{
	FTPCMD *command;
	char *parameters;
} INPUTLINE;

typedef struct configdata
{
	int maxthreads;      /* Max number of threads program can execute */
	int *listenport;     /* port to listen to for connections */
	unsigned int *listenips;	     /* Ip to bind to */
	int *listenfds;	     /* file descriptors we are listening to */
	int listenportlen;
	
	int loglevel;        /* Logging level */

	char* logfile;       /* file to log to */
	char* accessdir;    /* dir to get user access permissions from */
	
	int iplistlen;	     /* Length of iplist array */
	unsigned int *iplist;	     
	                     /* array of valid ips to accept */
	unsigned int *netmasklist;    
	                     /* list of netmasks corresponing to the ips */
	
	int timeout;	     /* timeout for ftp server */
	
	pid_t parentpid;       /* the pid to signal when an fatal error occurs */

	int logout;	     /* File to log out too */
	int umask;	     /* umask to default to */
	
	int rootmode;        /* are we in root mode or not */
	uid_t asuid;	     /* change to this user after listening to port */
	gid_t asgid;
	char *prelogondump;  /* file to dump before the user logs in */
	char *pidfile;       /* file to store this program's pid in */
	char *email;	     /* email address specified it config file */
	char *username;	     /* the username muddleftpd will run as */
	int logincount;      /* number of times a user may login before the
			        connection gets terminated */
	int authwait;	     /* time user must if login is incorrect */
	uid_t nobodyuid;     /* uid for user nobody */
	gid_t nobodygid;     /* gid for user nobody */
} CONFIGDATA;

/* GLOBAL VARIABLES */

CONFIGDATA *config;	  /* configuration data stored here for easy access */
THREADSTAT *pnums;	  /* process numbers and pids */

/* module memory.c */

extern void *mallocwrapper(int size);
extern void reallocwrapper(int size, void **inarea);
extern char *strdupwrapper(char *s);
extern pid_t forkwrapper(void);
extern void freewrapper(void *tofree);
extern char *safe_snprintf(char *format, ...);

/* module string.c */

extern int strchrcount(char *string, char tok);
extern void strclear(char *string);
extern char *strdupifnull(char *in, char *data);
extern char *gethostname2(void);

/* module error.c */

#define ERRORMSGFATAL(x)	errormsgfatal(x, __FILE__, __LINE__)
#define ERRORMSG(x)		errormsg(x, __FILE__, __LINE__)
extern void errormsg(char *errmessage, char *file, int line);
extern void errormsgfatal(char *errmessage, char *file, int line);

/* module shutdown.c */

extern void ftpd_shutdown(void);

/* module init.c */

extern void ftpd_init(char *filename);

/* module socket.c */

extern int getnetworkint(char *server, unsigned int *out);
extern int listenport(int port, unsigned int bindip, int maxconnect);
extern int listenparrelelport(int fd, int *port, unsigned int *ip, int maxconnect);
extern int get_conn(int socketin, unsigned int *addr);
extern int conn_server_nonblocking(unsigned int ip, int port, int localport);
extern int conn_server(unsigned int ip, int port);
extern void listen_close(int socketin);
extern int pipe2way(int fds[2]);
extern int socklock(int fd);
extern int sockunlock(int fd);
extern int nagleoff(int fd);
extern void socket_flush_wait(int fd, int timeout);

/* module reentrant.c */

extern char *ctime_r2(int timeint);
extern char *crypt_r2(char *key, char *salt);
extern char *ftp_time(int timeint);

/* module logger.c */

extern void log_init(void);
extern void log_shutdown(void);
extern void log_addentry(int type, unsigned int ip, char *username, char *desc);

/* module procnum.c */

extern void pnums_init(void);
extern int pnums_getnum_start(void);
extern void pnums_getnum_finish(int processnum, pid_t pid, int ipaddress);
extern void pnums_free_bypid(pid_t pid);
extern void pnums_dofree(void);
extern void pnums_print(FILE *of);
extern void pnums_setusername(int threadnum, char *username);
extern int pnums_countuser(char *username);
extern void pnums_killchildren(void);
extern void pnums_shutdown(void);
extern void pnums_setop(char *newop, int num);
extern int pnums_numusers(void);

/* module proxy.c */

extern int inport_bind(int portid, unsigned int ip);
extern int inport_getconn(SELECTER *mainsel, void *vincon);

/* module checkip.c */

extern int user_allowed(int ip);

/* module cmd.c */

extern char *getkeyword(char **);
extern void cmd_split(FTPSTATE *peer, INPUTLINE *cmd, char *, FTPCMD *, int log);

/* module ftp.c */

extern int ftp_run(FTPSTATE *peer, INPUTLINE *cmd, char *token);
extern int ftp_chmod(FTPSTATE *peer, char *filename, int mode);
extern int ftp_dumper(FTPSTATE *peer, NEWFILE *infile, int number, char *fmessage, int dotokens);
extern void reporterror(FTPSTATE *peer, char *filename, int error);
extern int ftp_dohelp(FTPSTATE *peer, FTPCMD *ftpcmds, char *params);

/* module auth.c */

extern char *setuseropts(FTPSTATE *peer, char *password);

/* module mudauth.c */

extern char *getmudpasswd(char *userfile);

/* module dir.c */

extern void dir_tovirtual(FTPSTATE *peer, char *normaldir);
extern void dir_toreal(FTPSTATE *peer, char *virdir);
extern void dir_combine(FTPSTATE *peer, char **pwd, char *newdir);

/* module file.c */

extern int checkdir(char *dir);
extern int checkfile(char *file);
extern void convertperms(char *permstr, int mode);
extern int checkchdir(FTPSTATE *peer, char *dir);
extern int file_creat(FTPSTATE *peer, char *filename);
extern int file_open(FTPSTATE *peer, char *filename, int flag);
extern NEWFILE *file_nfopen(FTPSTATE *peer, char *filename);
extern int file_glob(FTPSTATE *peer, char *filename, glob_t *gdata, int allfiles);
extern int file_unlink(FTPSTATE *peer, char *filename);
extern char *file_rename(FTPSTATE *peer, char *filename, char *filename2);
extern int file_rmdir(FTPSTATE *peer, char *filename);
extern int file_mkdir(FTPSTATE *peer, char *filename);
extern int file_stat(FTPSTATE *peer, char *filename, struct stat *output);
extern int file_chmod(FTPSTATE *peer, char *filename, int mode);
extern char *file_expand(FTPSTATE *peer, char *filename);

/* module acl.c */

extern void acl_clearcache(FTPSTATE *peer);
extern int check_acl(FTPSTATE *ftp, char *pathname, char perm);
extern void addaclstr(FTPSTATE *peer, char *accstr);
extern void acl_makecache(FTPSTATE *peer);

/* module ftpstate.c */

extern void ftpstate_init(FTPSTATE *ftp, int fd, int ip, int threadnum, int portnum);
extern void ftpstate_dest(FTPSTATE *ftp);

/* module ftpout.c */

extern int ftp_write(FTPSTATE *peer, int messnum, char *format, ...);

/* module bufread.c */

extern int readcmd(FTPSTATE *peer);
extern int popcmd(FTPSTATE *peer, char **outstr);

/* module ftpstat.c */

extern int ftp_stat(FTPSTATE *peer, char *param);

/* module ftpsite.c */

extern int ftp_site(FTPSTATE *peer, char *param);

/* module strcache.c */

extern STRCACHE *strcache_new(void);
extern char *strcache_check(STRCACHE *cache, int num);
extern void strcache_add(STRCACHE *cache, int num, char *str);
extern void strcache_free(STRCACHE *cache);

/* module bufwrite.c */

extern OUTBUFF *obuff_new(int outsock);
extern int obuff_writes(OUTBUFF *bout, char *output);
extern int obuff_write(OUTBUFF *bout, void *output, int outlen); 
extern int obuff_flush(OUTBUFF *bout);
extern void obuff_free(OUTBUFF *bout);

/* module config.c */

char *strsplit(char *, char *, char *);
void substitute(char **inputstr, char *strin, char *strout);
int loadconfigfile(char *filename, int (* confighandler)(char *, char *, int, void *), void *configdata);

/* module select.c */

extern SELECTER *select_new(void);
extern void select_addread(SELECTER *sel, int newport, int (* proc)(SELECTER *, void *), void *);
extern void select_addwrite(SELECTER *sel, int newport, int (* proc)(SELECTER *, void *), void *);
extern void select_takeread(SELECTER *sel, int deadport);
extern void select_takewrite(SELECTER *sel, int deadport);
extern int select_do(SELECTER *sel, int *signum, int timeout);
extern void select_addfd(SELECTER *sel, int newport);
extern void select_delfd(SELECTER *sel, int deadport);
extern void select_shutdown(SELECTER *sel);

/* module dataport.c */

extern int startdatasocket(FTPSTATE *peer, int filefd, int (* asciifilter)(char *, int *));
extern void closedatasocket(FTPSTATE *peer);
extern void abortdatasocket(FTPSTATE *peer);

/* module newfile.c */

extern int readbuffer(NEWFILE *file, int maxsize);
extern NEWFILE *nfopen(char *filename);
extern NEWFILE *nfdopen(int fd);
extern char *nfgetcs(NEWFILE *file, char testchar);
extern void nfclose(NEWFILE *file);

/* module ftplist.c */

extern int ftplist_parseflags(char *params);
extern int ftp_lister(FTPSTATE *peer, char *parm, int nlist, int params);
extern char *listfiles(FTPSTATE *peer, char *pattern, int nlist, int *usedstring, int params);
extern char *listmakepattern(FTPSTATE *peer, char *parm, int flags);

/* module shmem.c */

extern void shmem_lock(int id);
extern void shmem_unlock(int id);
extern void *shmem_malloc(int size);
extern void shmem_free(void *newpos);
extern void shmem_init(char *configfile);
extern void shmem_stats(void);
extern void shmem_finish(void);

/* module tokset.c */

extern TOKENSET *tokenset_new(void);
extern void tokenset_settoken(TOKENSET *tset, unsigned char token, char *data);
extern void tokenset_deltoken(TOKENSET *tset, unsigned char token);
extern char *tokenset_apply(TOKENSET *tok, char *inputstr);
extern void tokenset_finish(TOKENSET *tset);

#if (HAVE_SETFSUID == 1)
extern int setfsuid(uid_t uid);
#endif
#if (HAVE_SETFSGID == 1)
extern int setfsgid(uid_t gid);
#endif
