/* auth.c Handle user authentication and setup of user options

   Copyright (C) 1999 Beau Kuiper

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include "ftpd.h"
#include "auth.h"

PERMINFO authtypes[] = {
	{ "unix", &unixauth_commands },
	{ "anonymous", &anonauth_commands },
	{ "internal", &internalauth_commands },
	{ "disabled", &disableauth_commands },
#ifdef HAVE_PAM_START
	{ "pam", &pamauth_commands },
#endif
#ifdef MUDAUTH
	{ "mud", &mudauth_commands },
#endif
	{ NULL, NULL },
};

struct configparms
{
	char *userjail;
	char *groupname;
	char *authmethod;
	char *authparams;
	char *timeout;
	char *uid;
	char *gid;
	char *homedir;
	char *basedir;
	char *welcomefile;
	char *umask;
	char *chmoding;
	char *authuser;
	char *maxusers;
	char *cwddump;
	char *busydump;
	char *quitdump;
	PERMSTRUCT *authproc;
	void *authhandle;
	char **acclist;
	int acclistcount;
};

/* check an encrypted password against a plain text password */

int chkpassword(char *encrypass, char *password)
{
	char *pass2;
	int result = FALSE;
	
	if (strlen(encrypass) != 13)
		return(FALSE);
	
	pass2 = crypt(password, encrypass);
	if (strcmp(pass2, encrypass) == 0)
		result = TRUE;

	/* destroy the passwords so it cannot be seen if muddleftpd crashes! */
	memset(password, 0, strlen(password));
	memset(pass2, 0, strlen(pass2));
	
	return(result);
}

/* Make sure the user isn't trying to exploit the nature of the ftp server */

int checkexploits(FTPSTATE *peer)
{
	if (strchr(peer->username, '/'))
		return(TRUE);
	if (strlen(peer->username) > MAXUSERLEN)
		return(TRUE);
	return(FALSE);
}

/* initialize the user config struture */

void clearauth(struct configparms *cfpeer)
{
	/* just set everything to null by setting the entire structure to 0 */
	/* This may break on some archetectures :( */	
	memset(cfpeer, 0, sizeof(struct configparms));
}

int userconfigset(char *cmd, char *data, int line, void *d)
{
	struct configparms *c = (struct configparms *)d;

	if (strcasecmp(cmd, "group") == 0)	
		c->groupname = strdupifnull(c->groupname, data);
	else if (strcasecmp(cmd, "rootdir") == 0)
		c->basedir = strdupifnull(c->basedir, data);	
	else if (strcasecmp(cmd, "homedir") == 0)
		c->homedir = strdupifnull(c->homedir, data);
	else if (strcasecmp(cmd, "welcome") == 0)
		c->welcomefile = strdupifnull(c->welcomefile, data);
	else if (strcasecmp(cmd, "access") == 0)
	{
		c->acclistcount++;
		reallocwrapper(c->acclistcount * sizeof(char *), (void *)&(c->acclist));
		c->acclist[c->acclistcount - 1] = (char *)strdupwrapper(data);
	}
	else if (strcasecmp(cmd, "umask") == 0)
 		c->umask = strdupifnull(c->umask, data);
	else if (strcasecmp(cmd, "authuser") == 0)
		c->authuser = strdupifnull(c->authuser, data);
	else if (strcasecmp(cmd, "maxconnect") == 0)
		c->maxusers = strdupifnull(c->maxusers, data);								
	else if (strcasecmp(cmd, "timeout") == 0)
		c->timeout = strdupifnull(c->timeout, data);
	else if (strcasecmp(cmd, "chmoding") == 0)
		c->chmoding = strdupifnull(c->chmoding, data);
	else if (strcasecmp(cmd, "authmethod") == 0)
		c->authmethod = strdupifnull(c->authmethod, data);
	else if (strcasecmp(cmd, "authparams") == 0)
		c->authparams = strdupifnull(c->authparams, data);
	else if (strcasecmp(cmd, "uid") == 0)
		c->uid = strdupifnull(c->uid, data);
	else if (strcasecmp(cmd, "gid") == 0)
		c->gid = strdupifnull(c->gid, data);
	else if (strcasecmp(cmd, "cddump") == 0)
		c->cwddump = strdupifnull(c->cwddump, data);
	else if (strcasecmp(cmd, "busydump") == 0)
		c->busydump = strdupifnull(c->busydump, data);
	else if (strcasecmp(cmd, "quitdump") == 0)
		c->quitdump = strdupifnull(c->quitdump, data);
	else if (strcasecmp(cmd, "userjail") == 0)
		c->userjail = strdupifnull(c->userjail, data);
	return(TRUE);
}

void configsetauthmethod(struct configparms *c)
{
	int count = 0;
	c->authproc = NULL;
	
	while(authtypes[count].authname != NULL)
	{
		if (strcasecmp(c->authmethod, authtypes[count].authname) == 0)
		{
			c->authproc = authtypes[count].authstruct;
			return;
		}
		count++;
	}
}

void configsetdefaults(struct configparms *c)
{
	c->groupname = strdupifnull(c->groupname, "<none>");
	if (c->timeout == NULL)
		c->timeout = safe_snprintf("%d", config->timeout);
		 
	c->uid = strdupifnull(c->uid, "%u");
	c->gid = strdupifnull(c->gid, "%g");
	c->homedir = strdupifnull(c->homedir, "%h");
	c->basedir = strdupifnull(c->basedir, "%r");

	if (c->umask == NULL)
		c->umask = safe_snprintf("%o", config->umask);

	c->chmoding = strdupifnull(c->chmoding, "0");
	c->userjail = strdupifnull(c->userjail, "0");

	if (c->maxusers == NULL)
		c->maxusers = safe_snprintf("%d", config->maxthreads);
	c->authuser = strdupifnull(c->authuser, "%U");
}

void expandconfig(struct configparms *c, char *username)
{
	int count;
	TOKENSET *authtoks = tokenset_new();

	/* Ok. the following expansions are done 
 		%u	uid in password file 
		%g	gid in password file 
		%U	username 
		%h	homedir in password file 
		%r	rootdir from authmethod
	*/

	tokenset_settoken(authtoks, 'U', strdupwrapper(username));	
	
	/* Expand unixuser so we can get password info */
	c->authuser = tokenset_apply(authtoks, c->authuser);
	c->authparams = tokenset_apply(authtoks, c->authparams);
	
	/* map the authmethod parameter to get the procedures we want */
	
	configsetauthmethod(c);
	
	if (c->authproc)	/* now we can talk to the it */
	{
		c->authhandle = c->authproc->gethandle(c->authuser, c->authparams);
		if (c->authhandle != NULL)
		{
			if (c->authproc->getuseruid)
				tokenset_settoken(authtoks, 'u', safe_snprintf("%d", c->authproc->getuseruid(c->authhandle)));
			if (c->authproc->getusergid)
				tokenset_settoken(authtoks, 'g', safe_snprintf("%d", c->authproc->getusergid(c->authhandle)));
			if (c->authproc->gethomedir)
				tokenset_settoken(authtoks, 'h', strdupwrapper(c->authproc->gethomedir(c->authhandle)));
			if (c->authproc->getrootdir)
				tokenset_settoken(authtoks, 'r', strdupwrapper(c->authproc->getrootdir(c->authhandle)));
		}
		else
		{
			/* if we can't find the unix user, then don't continue */
			tokenset_finish(authtoks);
			return;
		}
	}
	else
	{
		tokenset_finish(authtoks);
		return;
	}
	
	/* Expand the rest of the options */

	c->timeout = tokenset_apply(authtoks, c->timeout);
	c->uid = tokenset_apply(authtoks, c->uid);
	c->gid = tokenset_apply(authtoks, c->gid);
	c->homedir = tokenset_apply(authtoks, c->homedir);
	c->basedir = tokenset_apply(authtoks, c->basedir);
	c->welcomefile = tokenset_apply(authtoks, c->welcomefile);
	c->umask = tokenset_apply(authtoks, c->umask);
	c->chmoding = tokenset_apply(authtoks, c->chmoding);
	c->userjail = tokenset_apply(authtoks, c->userjail);
	c->maxusers = tokenset_apply(authtoks, c->maxusers);
	c->cwddump = tokenset_apply(authtoks, c->cwddump);
	c->busydump = tokenset_apply(authtoks, c->busydump);
	c->quitdump = tokenset_apply(authtoks, c->quitdump);
	
	for(count = 0; count < c->acclistcount; count++)
		c->acclist[count] = tokenset_apply(authtoks, c->acclist[count]);
	
	tokenset_finish(authtoks);
}	

int checkforabsfile(char *filename, char *description, char *username)
{
	/* If no filename is specified, then no prob */
	
	if (!filename)
		return(TRUE);
		 
	if (filename[0] != '/')
	{
		char *errmsg = safe_snprintf("%s is not an absolute pathname (%s) for user %s", description, filename, username);
		log_addentry(MYLOG_INFO, 0, NULL, errmsg);
		freewrapper(errmsg);
		return(FALSE);
	}
	return(TRUE); 
}

int moveconfig(FTPSTATE *peer, struct configparms *c)
{
	int count, result = TRUE;
	int asgid, asuid;
	
	if (sscanf(c->timeout, "%d", &(peer->timeout)) != 1)
		peer->timeout = config->timeout;
	peer->maxtimeout = peer->timeout;
			
	if (sscanf(c->uid, "%d", &asuid) != 1)
		asuid = config->nobodyuid;
	if (asuid == 0)
		asuid = config->nobodyuid;
	peer->asuid = (uid_t)asuid;
	
	if (sscanf(c->gid, "%d", &asgid) != 1)
		asgid = config->nobodygid;
	if (asgid == 0)
		asgid = config->nobodygid;
	peer->asgid = (gid_t)asgid;
	
	if (peer->homedir)
		freewrapper(peer->homedir);
	peer->homedir = c->homedir;
	c->homedir = NULL;
	if (!checkforabsfile(peer->homedir, "homedir", peer->username))
		result = FALSE;
		  
	if (peer->basedir)
		freewrapper(peer->basedir);
	peer->basedir = c->basedir;
	c->basedir = NULL;
	if (!checkdir(peer->basedir))
	{
		char *errmsg = safe_snprintf("'%s' is not valid rootdir (does not exist) for user %s", peer->basedir, peer->username);
		log_addentry(MYLOG_INFO, 0, NULL, errmsg);
		freewrapper(errmsg);
		result = FALSE;
	}
	if (!checkforabsfile(peer->basedir, "rootdir", peer->username))
		result = FALSE;
		
	if (peer->logindump)
		freewrapper(peer->logindump);
	peer->logindump = c->welcomefile;
	c->welcomefile = NULL;
	if (!checkforabsfile(peer->logindump, "welcome", peer->username))
		result = FALSE;
	
	if (peer->cwddump)
		freewrapper(peer->cwddump);
	peer->cwddump = c->cwddump;
	c->cwddump = NULL;

	if (peer->busydump)
		freewrapper(peer->busydump);
	peer->busydump = c->busydump;
	c->busydump = NULL;
	if (!checkforabsfile(peer->busydump, "busydump", peer->username))
		result = FALSE;

	if (peer->quitdump)
		freewrapper(peer->quitdump);
	peer->quitdump = c->quitdump;
	c->quitdump = NULL;
	if (!checkforabsfile(peer->quitdump, "quitdump", peer->username))
		result = FALSE;
	
	if (sscanf(c->umask, "%o", &peer->umask) != 1)
		peer->umask = config->umask & 0777;
	
	if (sscanf(c->chmoding, "%d", &peer->chmodable) != 1)
		peer->chmodable = 0;

	if (sscanf(c->userjail, "%d", &peer->jailenabled) != 1)
		peer->jailenabled = 0;

	if (sscanf(c->maxusers, "%d", &peer->maxusers) != 1)
		peer->maxusers = config->maxthreads;
	
	if (peer->pwd)
		freewrapper(peer->pwd);
	peer->pwd = NULL;
	
	if (peer->renameoldname)
	{
		freewrapper(peer->renameoldname);
		peer->renameoldname = NULL;
	}
	
	peer->restartpos = 0;
	peer->binary = TRUE;
	peer->remoteport = peer->connport - 1;
	
	if (peer->passiveport)
		select_delfd(peer->sel, peer->passiveport);
	peer->passiveport = 0;	

	acl_clearcache(peer);
	acl_makecache(peer);
	
	for(count = 0; count < c->acclistcount; count++)
		addaclstr(peer, c->acclist[count]);

	return(result);
}

void freeconfigdata(struct configparms *c)
{
	int count;
	
	if (c->groupname) freewrapper(c->groupname);
	if (c->authmethod) freewrapper(c->authmethod);
	if (c->authparams) freewrapper(c->authparams);
	if (c->timeout) freewrapper(c->timeout);
	if (c->uid) freewrapper(c->uid);
	if (c->gid) freewrapper(c->gid);
	if (c->homedir) freewrapper(c->homedir);
	if (c->basedir) freewrapper(c->basedir);
	if (c->welcomefile) freewrapper(c->welcomefile);
	if (c->umask) freewrapper(c->umask);
	if (c->chmoding) freewrapper(c->chmoding);
	if (c->authuser) freewrapper(c->authuser);
	if (c->maxusers) freewrapper(c->maxusers);
	if (c->cwddump) freewrapper(c->cwddump);
	if (c->busydump) freewrapper(c->busydump);
	if (c->quitdump) freewrapper(c->quitdump);
	if (c->authhandle) c->authproc->freehandle(c->authhandle);	
	if (c->userjail) freewrapper(c->userjail);
	if (c->acclist)
	{
		for (count = 0; count < c->acclistcount; count++)
			freewrapper(c->acclist[count]);
		freewrapper(c->acclist);
	}
}
	
char *setuseropts(FTPSTATE *peer, char *password)
{
	char *groupname;
	char *userfilename;
	struct configparms userconfig;
	int error = TRUE;
	char *errmsg = NULL;
	
	peer->loggedin = FALSE;

	/* Check possible exploits here! */ 
	if (checkexploits(peer))
		return(strdupwrapper("Bad password"));
	
	/* Set all items in the FTPSTATE structure to not there */
	clearauth(&userconfig);
	
	chdir(config->accessdir);	
	userfilename = safe_snprintf("%s/user.%s", config->accessdir, peer->username);
	
	if (!checkfile(userfilename))
	{	
		freewrapper(userfilename);
		userfilename = safe_snprintf("%s/user.default", config->accessdir); 
	}
	
	if (checkfile(userfilename))
	{
		error = !loadconfigfile(userfilename, userconfigset, &userconfig);
		if ((!error) && (userconfig.groupname != NULL))
			if (strcasecmp(userconfig.groupname, "<none>") != 0)
			{
				freewrapper(userfilename);
				userfilename = safe_snprintf("%s/group.%s", config->accessdir, userconfig.groupname);
				error = !loadconfigfile(userfilename, userconfigset, &userconfig);  
			}
	}

	if (!error)
	{
		configsetdefaults(&userconfig);
		expandconfig(&userconfig, peer->username);
		if (userconfig.authhandle)
		{
			peer->loggedin = userconfig.authproc->chkpasswd(userconfig.authhandle, password, &errmsg);
			if (peer->loggedin)
				peer->loggedin = moveconfig(peer, &userconfig);
		}
	}
	
	if (error)
	{
		peer->loggedin = FALSE;
		groupname = safe_snprintf("Couldn't read authentication file for %s.", peer->username); 
		log_addentry(MYLOG_INFO, 0, NULL, groupname);
		freewrapper(groupname);
	}
	
	freeconfigdata(&userconfig);
	freewrapper(userfilename);
	if (peer->loggedin)
	{
		if (errmsg)
			freewrapper(errmsg);
		errmsg = NULL;
	}
	else
	{
		if (!errmsg)
			errmsg = safe_snprintf("Bad password");
	}
	return(errmsg);	
}
